// SPDX-License-Identifier: MIT
//
// Copyright IBM Corp. 2023

use clap::{Args, Parser, Subcommand, ValueHint};

use clap::ArgGroup;
// #[cfg(target_arch = "s390x")]
use pv::misc::stdout_possible_parser;
use pv::misc::CertificateOptions;
use pv::misc::{stdin_possible_parser, STDOUT};

/// Manage secrets for IBM Secure Execution guests.
///
/// It can be used to create and send Add Secret requests, list the added secrets and lock the
/// Secret Store.
#[derive(Parser, Debug)]
pub struct CliOptions {
    /// Provide more detailed output.
    #[arg(short='v', long, action = clap::ArgAction::Count, alias("V"))]
    pub verbose: u8,

    #[command(subcommand)]
    pub cmd: Command,
}

#[derive(Args, Debug, Default)]
#[command(group(ArgGroup::new("as-ext").args(["cck", "extension_secret"])),)]
pub struct CreateSecretOpt {
    #[command(flatten)]
    pub certificate_args: CertificateOptions,

    /// Specifies the header of the guest image.
    ///
    /// Can be a SE-image created by genprotimg or an extracted SE-header.
    /// The header must start at a page boundary.
    #[arg(long, value_name = "FILE", value_hint = ValueHint::FilePath, alias("se-hdr") )]
    pub hdr: String,

    /// Write the generated request to FILE.
    #[arg(short, long, value_name = "FILE", value_parser = stdout_possible_parser, value_hint = ValueHint::FilePath,)]
    pub output: String,

    /// Use the content of FILE as an extension secret.
    ///
    /// The file must be exactly 32 bytes long
    /// If this request is the first one all following request must have the same extension secret.
    /// Only makes sense if bit 1 of the secret control flags of the secure
    /// execution header is 0, otherwise the UV will reject the request.
    #[arg(long, value_name = "FILE", value_hint = ValueHint::FilePath,)]
    pub extension_secret: Option<String>,

    /// Use the content of FILE as the CCK to derive the extension secret.
    ///
    /// The file must contain exactly 32 byte of data.
    /// If the target guest was started with bit 1 of the secret control flag set,
    /// the Ultravisor will also derive the secret from the Customer Communication Key.
    /// Otherwise, UV will interpret the extension secret as a normal one.
    /// This should still work, if you use the same CCK for all requests.
    #[arg(long, value_name = "FILE", alias("comm-key"))]
    pub cck: Option<String>,

    /// Use `HEXSTRING` as the Configuration Unique Id.
    ///
    /// Must be a hex 128 byte unsigned big endian number string. Leading zeros must be provided.
    /// If specified, the value must match with the Config-UID from the Attestation result of that guest.
    /// If not specified, the CUID will be ignored by the Ultravisor during the verification of the request.
    #[arg(long, value_name = "HEXSTRING")]
    pub cuid_hex: Option<String>,

    /// Use the content of FILE as the Configuration Unique Id.
    ///
    /// The file must contain exactly 128 byte of data, a hex-string, or a yaml with a `cuid` entry.
    /// If specified, the value must match with the Config-UID from the Attestation result of that guest.
    /// If not specified, the CUID will be ignored by the Ultravisor during the verification of the request.
    #[arg(long, value_name = "FILE", conflicts_with("cuid_hex"), value_hint = ValueHint::FilePath,)]
    pub cuid: Option<String>,

    #[command(subcommand)]
    pub secret: AddSecretType,

    // FLAGS
    // each flag must conflict with `flags`
    // `flags` is hidden in the help menu
    /// Manually set the Add Secret Request flags.
    ///
    /// No validity checks made.
    #[arg(long, hide(true))]
    pub flags: Option<u64>,

    /// Disables host initiated dumping for the target guest.
    ///
    /// Dumping cannot be enabled again after this request was dispatched successfully.
    #[arg(long, conflicts_with("flags"))]
    pub disable_dump: bool,
}

#[derive(Subcommand, Debug, Default)]
pub enum AddSecretType {
    /// Use a Null-secret, containing no secret information.
    ///
    /// A Null-secret can be used to carry flags to the UV without having to provide an actual
    /// secret.
    #[default]
    Null,

    /// Use an Association-secret to connect an I/O device to a guest.
    ///
    /// An Association-secret is used to associate a trusted IO-device to a SE-guest.
    /// The `pvapconfig` tool provides for more information on Association-secrets.
    Association {
        /// String to identify the new secret.
        ///
        /// String will be hashed with SHA256 to form the id used by the UV.
        /// Actual secret is set with the `secret` argument.
        /// Name Id will be saved in `NAME.yaml` with white-spaces mapped to `_`.
        name: String,

        ///Print the hased name to stdout.
        ///
        ///The hasehd name will not be written to `NAME.yaml`
        #[arg(long)]
        stdout: bool,

        /// Path to read the secret from. Use a random secret if not specified.
        ///
        /// The secret will be stored in `NAME` if file is not used as input here
        /// (See `name` parameter).
        #[arg(long, value_name = "FILE", value_hint = ValueHint::FilePath,)]
        secret: Option<String>,
    },
}

// #[cfg(target_arch = "s390x")]
#[derive(Args, Debug)]
pub struct AddSecretOpt {
    /// Specify the request to be sent.
    #[arg(value_name = "FILE", value_parser = stdin_possible_parser, value_hint = ValueHint::FilePath,)]
    pub input: String,
}

// #[cfg(target_arch = "s390x")]
#[derive(Args, Debug)]
pub struct ListSecretOpt {
    /// Store the result in FILE
    #[arg(value_name = "FILE", default_value = STDOUT, value_parser = stdout_possible_parser, value_hint = ValueHint::FilePath,)]
    pub output: String,

    /// Use yaml format.
    ///
    /// Outputs the list in YAML format instead of the default human targeted one.
    #[arg(long)]
    pub yaml: bool,

    /// Use binary format.
    ///
    /// Outputs the list in the same binary format the Ultravisor uses.
    #[arg(long, conflicts_with = "yaml")]
    pub bin: bool,
}

#[derive(Subcommand, Debug)]
pub enum Command {
    /// Create a new Add-secret request.
    Create(Box<CreateSecretOpt>),

    /// Do an Add-secret request using a previously generated `Add Secret` request.
    ///
    /// Only available on s390x.
    // #[cfg(target_arch = "s390x")]
    Add(AddSecretOpt),
    // #[cfg(not(target_arch = "s390x"))]
    // Add,
    /// Lock the secret-store.
    ///
    /// After this command executed successfully, all `Add-secret requests` will fail.
    /// Only available on s390x.
    Lock,

    /// List all secrets the Ultravisor stores for this guest.
    ///
    /// Shows the id of all non-null secrets currently stored in the Ultravisor.
    /// Only available on s390x.
    // #[cfg(target_arch = "s390x")]
    List(ListSecretOpt),
    // #[cfg(not(target_arch = "s390x"))]
    // List,
    /// Print version information.
    Version,
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    #[rustfmt::skip]
    fn cli_args() {
        //Verify only that some arguments are optional, we do not want to test clap, only the
        //configuration
        let valid_args = [
            // #[cfg(target_arch = "s390x")]
            vec!["pvsecret", "lock"],
            vec!["pvsecret", "version"],
            // #[cfg(target_arch = "s390x")]
            vec!["pvsecret", "list"],
            // #[cfg(target_arch = "s390x")]
            vec!["pvsecret", "add", "abc"],
            vec!["pvsecret", "create", "-k", "abc", "--se-hdr", "abc", "-o", "abc", "--no-verify", "null"],
            vec!["pvsecret", "create", "-k", "abc", "--se-hdr", "abc", "-o", "abc", "--no-verify", "association", "name" ],
        ];
        // Test for the minimal amount of flags to yield an invalid combination
        let invalid_args = [
            vec!["pvsecret"],
            vec!["pvsecret", "list", "--yaml", "--bin"],
            vec!["pvsecret", "create", "--se-hdr", "abc", "-o", "abc", "--no-verify" ,"null"],
            vec!["pvsecret", "create", "-k", "abc", "-o", "abc", "--no-verify", "null"],
            vec!["pvsecret", "create", "-k", "abc", "--se-hdr", "abc", "--no-verify", "null"],
            vec!["pvsecret", "create", "-k", "abc", "--se-hdr", "abc", "-o", "abc", "null"],
            vec!["pvsecret", "create", "-k", "abc", "--se-hdr", "abc", "-o", "abc", "--cck", "abc", "--extension_secret", "abc", "--no-verify", "null"],
            vec!["pvsecret", "create", "-k", "abc", "--se-hdr", "abc", "-o", "abc", "--no-verify", "--flags", "0", "--disable-dump", "null"],
            vec!["pvsecret", "create", "-k", "abc", "--se-hdr", "abc", "-o", "abc", "--no-verify", "--cuid", "abc", "--cuid_hex", "9",  "null"],
            vec!["pvsecret", "create", "-k", "abc", "--se-hdr", "abc", "-o", "abc", "--no-verify", "association" ],
            ];
        for arg in valid_args {
            let res = CliOptions::try_parse_from(&arg);
            println!("arg: {arg:?}");
            println!("res: {res:?}");
            assert!(res.is_ok());
        }

        for arg in invalid_args {
            let res = CliOptions::try_parse_from(&arg);
            assert!(res.is_err());
        }
    }

    #[test]
    fn verify_cli() {
        use clap::CommandFactory;
        CliOptions::command().debug_assert()
    }
}
