// SPDX-License-Identifier: MIT
//
// Copyright IBM Corp. 2023

mod cli;
mod cmd;

use clap::CommandFactory;
use clap::Parser;
use cli::{CliOptions, Command};
use log::trace;
use pv::misc::PvLogger;
use std::process::ExitCode;
use utils::release_string;

static LOGGER: PvLogger = PvLogger;
static EXIT_LOGGER: u8 = 3;

//TODO activate all arch barriers
fn print_error(e: anyhow::Error, verbosity: u8) -> ExitCode {
    if verbosity > 0 {
        // Debug formatter also prints the whole error stack
        // So only print it when on verbose
        eprintln!("error: {e:?}")
    } else {
        eprintln!("error: {e}")
    };
    ExitCode::FAILURE
}

// #[cfg(not(target_arch = "s390x"))]
// fn not_supported() -> anyhow::Result<()> {
//     use anyhow::bail;
//
//     bail!("Command only available on s390x")
// }

fn main() -> ExitCode {
    let cli = match CliOptions::try_parse() {
        Ok(cli) => cli,
        Err(e) => {
            let ret = if e.use_stderr() {
                ExitCode::FAILURE
            } else {
                ExitCode::SUCCESS
            };
            //Ignore any errors during printing of the error
            let _ = e.format(&mut CliOptions::command()).print();
            return ret;
        }
    };

    // set up logger/std(out,err)
    if let Err(e) = LOGGER.start(cli.verbose) {
        //should(TM) never happen
        eprintln!("Logger error: {e:?}");
        return EXIT_LOGGER.into();
    }

    // NOTE trace verbosity is disabled in release builds
    trace!("Trace verbosity, may leak secrets to command-line");
    trace!("Options {cli:?}");

    // perform the command selected by the user
    let res = match &cli.cmd {
        // #[cfg(target_arch = "s390x")]
        Command::Add(opt) => cmd::add(opt),
        // #[cfg(target_arch = "s390x")]
        Command::List(opt) => cmd::list(opt),
        // #[cfg(target_arch = "s390x")]
        Command::Lock => cmd::lock(),

        // #[cfg(not(target_arch = "s390x"))]
        // Command::Add => not_supported(),
        // #[cfg(not(target_arch = "s390x"))]
        // Command::List => not_supported(),
        // #[cfg(not(target_arch = "s390x"))]
        // Command::Lock => not_supported(),
        Command::Create(opt) => cmd::create(opt),
        Command::Version => {
            println!(
                "{} version {}\nCopyright IBM Corp. 2023",
                env!("CARGO_PKG_NAME"),
                release_string!()
            );
            if cli.verbose > 0 {
                println!(
                    "\n{}-crate {}",
                    env!("CARGO_PKG_NAME"),
                    env!("CARGO_PKG_VERSION")
                );
                println!("{}", pv::crate_info());
            }
            Ok(())
        }
    };

    match res {
        Ok(_) => ExitCode::SUCCESS,
        Err(e) => print_error(e, cli.verbose),
    }
}
